// Copy constructor and operator = : None, use default ones.

template <typename Long> class Large
{
public:

	static const Large Max;
	static const Large DivMax;

	Large() :
	m_loWord(0),
	m_hiWord(0)
	{
	}

	Large(const unsigned int x) :
	m_loWord(Long(x)),
	m_hiWord(0)
	{
	}

	Large(const Long & x) :
	m_loWord(x),
	m_hiWord(0)
	{
	}

	Large(const Long & lo, const Long & hi) :
	m_loWord(lo),
	m_hiWord(hi)
	{
	}

	~Large() {}

	Large operator+(const Large & other) const
	{
		// ab + cd
		Long t1, t2, t3, t4;
		t1 = Long(m_loWord.lo()) + Long(other.m_loWord.lo());
		t2 = Long(m_loWord.hi()) + Long(other.m_loWord.hi()) + t1.hi();
		t3 = Long(m_hiWord.lo()) + Long(other.m_hiWord.lo()) + t2.hi();
		t4 = Long(m_hiWord.hi()) + Long(other.m_hiWord.hi()) + t3.hi();
		return Large(Long(t1.lo()) + Long(0, t2.lo()),
		             Long(t3.lo()) + Long(0, t4.lo()));
	}

	Large operator-(const Large & other) const
	{
		// a1b1c1d1 - a2b2c2d2
		Long t1(m_loWord.lo(), 1);
		Long t2(m_loWord.hi(), 1);
		Long t3(m_hiWord.lo(), 1);
		Long t4(m_hiWord.hi(), 1);
		t2.subtractOne();
		t3.subtractOne();
		t4.subtractOne();
		t1 = (t1 - Long(other.m_loWord.lo()));
		t2 = (t2 - Long(other.m_loWord.hi())) + t1.hi();
		t3 = (t3 - Long(other.m_hiWord.lo())) + t2.hi();
		t4 = (t4 - Long(other.m_hiWord.hi())) + t3.hi();
		return Large(Long(t1.lo()) + Long(0, t2.lo()),
		             Long(t3.lo()) + Long(0, t4.lo()));
	}

	Large operator*(const Large & other) const
	{
		const Long d1 = m_loWord.lo();
		const Long c1 = m_loWord.hi();
		const Long b1 = m_hiWord.lo();
		const Long a1 = m_hiWord.hi();
		const Long d2 = other.m_loWord.lo();
		const Long c2 = other.m_loWord.hi();
		const Long b2 = other.m_hiWord.lo();
		const Long a2 = other.m_hiWord.hi();
		// d1:
		const Long d1d2 = d1 * d2;
		const Long d1c2 = d1 * c2;
		const Long d1b2 = d1 * b2;
		const Long d1a2 = d1 * a2;
		// c1:
		const Long c1d2 = c1 * d2;
		const Long c1c2 = c1 * c2;
		const Long c1b2 = c1 * b2;
		// b1:
		const Long b1d2 = b1 * d2;
		const Long b1c2 = b1 * c2;
		// a1:
		const Long a1d2 = a1 * d2;
		// result:
		return
			// d1:
			Large(d1d2) + Large(Long(0, d1c2.lo()), Long(d1c2.hi(), 0)) + Large(0, d1b2) + Large(0, Long(0, d1a2.lo())) +
			// c1:
			Large(Long(0, c1d2.lo()), Long(c1d2.hi(), 0)) + Large(0, c1c2) + Large(0, Long(0, c1b2.lo())) +
			// b1:
			Large(0, b1d2) + Large(0, Long(0, b1c2.lo())) +
			// a1:
			Large(0, Long(0, a1d2.lo()));
	}

	void div(const Large & other, Large & result, Large & remainder) const
	{
		if (other.isZero())
		{
			// Cause division by zero:
			int t = 1 / int(other.loOne());
			printf("%d", t);
			return;
		}
		if (*this < other)
		{
			result = 0;
			remainder = *this;
			return;
		}
		if (*this == other)
		{
			result = 1;
			remainder = 0;
			return;
		}
		result = 0;
		remainder = *this;
		Large divisor = other;
		Large toAdd = 1;
		while (!divisor.hiOne())
		{
			divisor.shiftLeft(false);
			toAdd.shiftLeft(false);
		}
		while (!(remainder < other))
		{
			while (remainder < divisor)
			{
				Large t = divisor;
				t.shiftRight(false);
				if (t < other)
				{
					break;
				}
				divisor = t;
				toAdd.shiftRight(false);
			}
			result = (result + toAdd);
			remainder = (remainder - divisor);
		}
	}

	Large operator/(const Large & other) const
	{
		Large result;
		Large remainder;
		div(other, result, remainder);
		return result;
	}

	Large operator%(const Large & other) const
	{
		Large result;
		Large remainder;
		div(other, result, remainder);
		return remainder;
	}

	void subtractOne()
	{
		if (m_loWord == Long(unsigned(0)))
		{
			m_hiWord.subtractOne();
		}
		m_loWord.subtractOne();
	}

	void shiftLeft(bool loBit)
	{
		bool loHiOne = m_loWord.hiOne();
		m_hiWord.shiftLeft(loHiOne);
		m_loWord.shiftLeft(loBit);
	}

	void shiftRight(bool hiBit)
	{
		bool hiLoOne = m_hiWord.loOne();
		m_hiWord.shiftRight(hiBit);
		m_loWord.shiftRight(hiLoOne);
	}

	Large & shiftRight()
	{
		shiftRight(false);
		return *this;
	}

	bool hiOne() const
	{
		return m_hiWord.hiOne();
	}

	bool loOne() const
	{
		return m_loWord.loOne();
	}

	bool operator<(const Large & other) const
	{
		if (m_hiWord < other.m_hiWord)
		{
			return true;
		}
		if (m_hiWord == other.m_hiWord)
		{
			return m_loWord < other.m_loWord;
		}
		return false;
	}

	bool operator==(const Large & other) const
	{
		return (m_loWord == other.m_loWord) && (m_hiWord == other.m_hiWord);
	}

	const Long & lo() const
	{
		return m_loWord;
	}

	const Long & hi() const
	{
		return m_hiWord;
	}

	bool isZero() const
	{
		return m_hiWord.isZero() && m_loWord.isZero();
	}

private:
	Long m_loWord;
	Long m_hiWord;
};

template <typename Long> const Large<Long> Large<Long>::Max(Long::Max, Long::Max);
template <typename Long> const Large<Long> Large<Long>::DivMax(
	Large<Long>::Max - Large<Long>(Large<Long>::Max).shiftRight());

template <> class Large <unsigned long>
{
private:

public:

	typedef unsigned short Short;
	typedef unsigned long Long;
	typedef unsigned long long LargeType;

	static const Large Max;

	Large() :
	m_number(0)
	{
	}

	Large(const LargeType & x) :
	m_number(x)
	{
	}

	Large(const Long lo, const Long hi) :
	m_number((LargeType(hi) << 32) | LargeType(lo))
	{
	}

	Large operator+(const Large & other) const
	{
		return Large(LargeType(m_number) + LargeType(other.m_number));
	}

	Large operator-(const Large & other) const
	{
		return Large(LargeType(m_number) - LargeType(other.m_number));
	}

	Large operator*(const Large & other) const
	{
		return Large(LargeType(m_number) * LargeType(other.m_number));
	}

	void div(const Large & other, Large & result, Large & remainder) const
	{
		result = Large(m_number / other.m_number);
		remainder = Large(m_number % other.m_number);
	}

	Large operator/(const Large & other) const
	{
		return Large(LargeType(m_number) / LargeType(other.m_number));
	}

	Large operator%(const Large & other) const
	{
		return Large(LargeType(m_number) % LargeType(other.m_number));
	}

	void subtractOne()
	{
		m_number -= 1;
	}

	void shiftLeft(bool loBit)
	{
		m_number = (m_number << 1) | LargeType(loBit);
	}

	void shiftRight(bool hiBit)
	{
		m_number = (m_number >> 1) | (LargeType(hiBit) << 63);
	}

	bool hiOne() const
	{
		return (m_number >> 63) == 1;
	}

	bool loOne() const
	{
		return (m_number & 1) == 1;
	}

	bool operator<(const Large & other) const
	{
		return m_number < other.m_number;
	}

	bool operator==(const Large & other) const
	{
		return m_number == other.m_number;
	}

	const Long lo() const
	{
		return Long(m_number & LargeType(0xffffffff));
	}

	const Long hi() const
	{
		return Long(m_number >> 32);
	}

	bool isZero() const
	{
		return m_number == 0;
	}

private:
	LargeType m_number;
};

const Large<unsigned long> Large<unsigned long>::Max((LargeType(0xffffffff) << 32) | LargeType(0xffffffff));

template <const size_t BitCount> struct Bits
{
	typedef Large<typename Bits<BitCount / 2>::Type> Type;
};

template <> struct Bits<64>
{
	typedef Large<unsigned long> Type;
};

typedef Bits<  64>::Type Large64;
typedef Bits< 128>::Type Large128;
typedef Bits< 256>::Type Large256;
typedef Bits< 512>::Type Large512;
typedef Bits<1024>::Type Large1024;
typedef Bits<2048>::Type Large2048;
